# -*- coding: utf-8 -*-

import sys

PY3 = sys.version_info[0] == 3

if PY3:
    from urllib.parse import quote, urlencode

    def b(s):
        if isinstance(s, str):
            return s.encode('utf-8')
        return s

    builtin_str = str
    str = str
    bytes = bytes
else:
    from urllib import quote, urlencode

    def b(s):
        return s

    builtin_str = str
    str = unicode  # noqa
    bytes = str

__all__ = [
    'quote', 'urlencode'
]



# -*- coding: utf-8 -*-
import os

from .modules.sign import make_rest_signature,\
    make_content_md5, encode_msg
from .modules.exception import UpYunClientException
from .modules.compat import b, str, quote, urlencode, builtin_str
from .modules.httpipe import cur_dt


def get_fileobj_size(fileobj):
    try:
        if hasattr(fileobj, 'fileno'):
            return os.fstat(fileobj.fileno()).st_size
    except IOError:
        pass

    return len(fileobj.getvalue())


class UploadObject(object):
    def __init__(self, fileobj, chunksize=None, handler=None, params=None):
        self.fileobj = fileobj
        self.chunksize = chunksize
        self.totalsize = get_fileobj_size(fileobj)
        self.readsofar = 0
        if handler:
            self.hdr = handler(self.totalsize, params)

    def __iter__(self):
        return self

    def __next__(self):
        chunk = self.fileobj.read(self.chunksize)
        if chunk and self.hdr:
            self.readsofar += len(chunk)
            if self.readsofar != self.totalsize:
                self.hdr.update(self.readsofar)
            else:
                self.hdr.finish()
        return chunk

    def __len__(self):
        return self.totalsize

    def read(self, size=-1):
        return self.__next__()


class UpYunRest(object):
    def __init__(self, bucket, username, password,
                 endpoint, chunksize, hp):
        self.bucket = bucket
        self.username = username
        self.password = password
        self.chunksize = chunksize
        self.endpoint = endpoint
        self.hp = hp
        self.host = None

    # --- public API
    def usage(self, key):
        res = self.__do_http_request('GET', key, args='?usage')
        return str(int(res))

    def put(self, key, value, checksum, headers, handler, params, secret):
        '''
        >>> with open('foo.png', 'rb') as f:
        >>>    res = up.put('/path/to/bar.png', f, checksum=False,
        >>>                 headers={'x-gmkerl-rotate': '180'}})
        '''
        if headers is None:
            headers = {}

        if isinstance(value, str):
            value = b(value)

        if checksum is True:
            headers['Content-MD5'] = make_content_md5(value, self.chunksize)

        if secret:
            headers['Content-Secret'] = secret

        if handler and hasattr(value, 'fileno'):
            value = UploadObject(value, chunksize=self.chunksize,
                                 handler=handler, params=params)

        h = self.__do_http_request('PUT', key, value, headers)
        return self.__get_meta_headers(h)

    def get(self, key, value, handler, params):
        '''
        >>> with open('bar.png', 'wb') as f:
        >>>    up.get('/path/to/bar.png', f)
        '''
        return self.__do_http_request('GET', key, of=value, stream=True,
                                      handler=handler, params=params)

    def delete(self, key):
        self.__do_http_request('DELETE', key)

    def mkdir(self, key):
        headers = {'Folder': 'true'}
        self.__do_http_request('POST', key, headers=headers)

    def getlist(self, key):
        content = self.__do_http_request('GET', key)
        if content == '':
            return []
        items = content.split('\n')
        return [dict(zip(['name', 'type', 'size', 'time'],
                x.split('\t'))) for x in items]

    def getinfo(self, key):
        h = self.__do_http_request('HEAD', key)
        return self.__get_meta_headers(h)

    def purge(self, keys, domain):
        domain = domain or '%s.b0.upaiyun.com' % (self.bucket)
        if isinstance(keys, builtin_str):
            keys = [keys]
        if isinstance(keys, list):
            urlfmt = 'http://%s/%s'
            urlstr = '\n'.join([urlfmt % (domain, k if k[0] != '/' else k[1:])
                                for k in keys]) + '\n'
        else:
            raise UpYunClientException('keys type error')

        method = 'POST'
        host = 'purge.upyun.com'
        uri = '/purge/'
        params = urlencode({'purge': urlstr})
        headers = {'Content-Type': 'application/x-www-form-urlencoded',
                   'Accept': 'application/json'}
        self.__set_auth_headers(urlstr, headers=headers)

        resp = self.hp.do_http_pipe(method, host, uri,
                                    value=params, headers=headers)
        content = self.__handle_resp(resp, method, uri=uri)
        invalid_urls = content['invalid_domain_of_url']
        return [k[7 + len(domain):] for k in invalid_urls if k]

    # --- private API
    def __do_http_request(self, method, key,
                          value=None, headers=None, of=None, args='',
                          stream=False, handler=None, params=None):
        _uri = '/%s/%s' % (self.bucket, key if key[0] != '/' else key[1:])
        uri = '%s%s' % (quote(encode_msg(_uri), safe='~/'), args)

        if headers is None:
            headers = {}

        if self.host:
            headers["HOST"] = self.host

        length = 0
        if hasattr(value, '__len__'):
            length = len(value)
            headers['Content-Length'] = str(length)
        elif hasattr(value, 'fileno'):
            length = get_fileobj_size(value)
            headers['Content-Length'] = str(length)
            # [ugly]:compatible with newest requests feature
            # force the stream upload with empty file to normal upload
            if not length:
                value = ''
        elif value is not None:
            raise UpYunClientException('object type error')

        self.__set_auth_headers(uri, method, length, headers)

        resp = self.hp.do_http_pipe(method, self.endpoint, uri,
                                    value, headers, stream)
        return self.__handle_resp(resp, method, of, handler, params)

    def __handle_resp(self, resp, method=None, of=None,
                      handler=None, params=None, uri=None):
        content = None
        try:
            if method == 'GET' and of:
                readsofar = 0
                try:
                    totalsize = int(resp.headers['content-length'])
                except (KeyError, TypeError):
                    totalsize = 0

                hdr = None
                if handler and totalsize > 0:
                    hdr = handler(totalsize, params)

                for chunk in resp.iter_content(self.chunksize):
                    if chunk and hdr:
                        readsofar += len(chunk)
                        if readsofar != totalsize:
                            hdr.update(readsofar)
                        else:
                            hdr.finish()
                    if not chunk:
                        break
                    of.write(chunk)
            elif method == 'GET':
                content = resp.text
            elif method == 'PUT' or method == 'HEAD':
                content = resp.headers.items()
            elif method == 'POST' and uri == '/purge/':
                content = resp.json()
        except Exception as e:
            raise UpYunClientException(e)
        return content

    def __get_meta_headers(self, headers):
        return dict((k[8:].lower(), v) for k, v in headers
                    if k[:8].lower() == 'x-upyun-' and
                    k[8:].lower() != 'uuid' and
                    k[8:].lower() != 'cluster')

    def __set_auth_headers(self, playload,
                           method=None, length=0, headers=None):
        if headers is None:
            headers = []

        dt = cur_dt()
        signature = make_rest_signature(self.bucket, self.username,
                                        self.password, method, playload,
                                        dt, length)

        headers['Authorization'] = signature
        return headers



# -*- coding: utf-8 -*-
import hashlib
import json
import os

from .rest import UpYunRest
from .form import FormUpload
from .multi import Multipart
from .av import AvPretreatment

from .modules.httpipe import UpYunHttp
from .modules.exception import UpYunClientException
from .modules.compat import b, builtin_str
from .modules.sign import make_content_md5, encode_msg
from .modules.check import has_object

__version__ = '2.3.6'

ED_LIST = ('v%d.api.upyun.com' % ed for ed in range(4))
ED_AUTO, ED_TELECOM, ED_CNC, ED_CTT = ED_LIST

DEFAULT_CHUNKSIZE = 8192


class UpYun(object):
    def __init__(self, bucket, username=None, password=None, secret=None,
                 timeout=None, endpoint=None, chunksize=None, debug=False,
                 read_timeout=None):
        super(UpYun, self).__init__()
        self.bucket = bucket or os.getenv('UPYUN_BUCKET')
        self.username = username or os.getenv('UPYUN_USERNAME')
        password = password or os.getenv('UPYUN_PASSWORD')
        self.password = (hashlib.md5(b(password)).hexdigest()
                         if password else None)
        self.endpoint = endpoint or ED_AUTO
        self.chunksize = chunksize or DEFAULT_CHUNKSIZE
        self.secret = secret or os.getenv('UPYUN_SECRET')
        self.timeout = timeout or 60
        if read_timeout is not None:
            self.requests_timeout = (self.timeout, read_timeout)
        else:
            self.requests_timeout = self.timeout
        self.hp = UpYunHttp(self.requests_timeout, debug)

        if self.username and self.password:
            self.up_rest = UpYunRest(self.bucket, self.username,
                                     self.password, self.endpoint,
                                     self.chunksize, self.hp)
            self.av = AvPretreatment(self.bucket, self.username,
                                     self.password, self.chunksize,
                                     self.hp)
        if self.secret:
            self.up_multi = Multipart(self.bucket, self.secret,
                                      self.endpoint, self.hp)
            self.up_form = FormUpload(self.bucket, self.secret,
                                      self.endpoint, self.hp)

        if debug:
            self.__init_debug_log(bucket=bucket, username=username,
                                  password=password, secret=secret,
                                  timeout=timeout, endpoint=endpoint,
                                  chunksize=chunksize, debug=debug)

    def __init_debug_log(self, **kwargs):
        with open('debug.log', 'w') as f:
            f.write('### Running in debug mode ###\n\n\n')
            f.write('## Initial params ##\n\n')
            f.write('\n'.join(map(lambda kv: '%s: %s'
                              % (kv[0], kv[1]), kwargs.items())))

    @has_object('up_rest')
    def set_endpoint(self, endpoint, host=None):
        self.up_rest.endpoint = endpoint
        self.up_rest.host = host or ED_AUTO

    # --- public rest API
    @has_object('up_rest')
    def usage(self, key='/'):
        return self.up_rest.usage(key)

    def put(self, key, value, checksum=False, headers=None,
            handler=None, params=None, secret=None,
            multipart=False, block_size=None, form=False,
            expiration=None, **kwargs):
        if (multipart or form) and not self.secret:
            raise UpYunClientException('You have to specify form secret with '
                                       'multipart upload method')

        # - priority: rest > form > multipart
        if form and hasattr(value, 'fileno'):
            return self.up_form.upload(key, value, expiration, **kwargs)
        if multipart and hasattr(value, 'fileno'):
            return self.up_multi.upload(key, value,
                                        block_size, expiration, **kwargs)
        return self.up_rest.put(key, value, checksum,
                                headers, handler, params, secret)

    @has_object('up_rest')
    def get(self, key, value=None, handler=None, params=None):
        return self.up_rest.get(key, value, handler, params)

    @has_object('up_rest')
    def delete(self, key):
        self.up_rest.delete(key)

    @has_object('up_rest')
    def mkdir(self, key):
        self.up_rest.mkdir(key)

    @has_object('up_rest')
    def getlist(self, key='/'):
        return self.up_rest.getlist(key)

    @has_object('up_rest')
    def getinfo(self, key):
        return self.up_rest.getinfo(key)

    @has_object('up_rest')
    def purge(self, keys, domain=None):
        return self.up_rest.purge(keys, domain)

    # --- video pretreatment API
    @has_object('av')
    def pretreat(self, tasks, source, notify_url=''):
        return self.av.pretreat(tasks, source, notify_url)

    @has_object('av')
    def status(self, taskids):
        return self.av.status(taskids)

    @has_object('av')
    def verify_tasks(self, value):
        return self.av.verify_tasks(value)

    # --- depress task
    @has_object('av')
    def depress(self, tasks, notify_url):
        for task in tasks:
            save_as = task.get('save_as')
            sources = task.get('sources')
            for key in (sources, save_as):
                if type(key) != str or key == '':
                    raise UpYunClientException('Given not correct %s '
                                               'in task' % key)
        return self.av.pretreat(tasks, 'upyun', notify_url, 'depress')

    # --- compress task
    @has_object('av')
    def compress(self, tasks, notify_url):
        for task in tasks:
            save_as = task.get('save_as')
            sources = task.get('sources')
            if type(save_as) != str or save_as == '':
                raise UpYunClientException('Given not correct save_as in task')
            if type(sources) != list or len(sources) == 0:
                raise UpYunClientException('Given not correct sources in task')
        return self.av.pretreat(tasks, 'upyun', notify_url, 'compress')

    @has_object('av')
    def put_tasks(self, tasks, notify_url, app_name):
        return self.av.pretreat(tasks, '', notify_url, app_name)


# --- no use yet, need developing
def verify_put_sign(value, secret):
    KEYS = ['code', 'message', 'url', 'time',
            'form_api_secret', 'ext-param', ]
    data = []

    if not isinstance(value, dict):
        value = json.loads(value)
    if 'no-sign' in value:
        value['form_api_secret'] = ''
        sign = value['no-sign']
    else:
        value['form_api_secret'] = secret
        sign = value['sign']
    for k in KEYS:
        if k == 'url':
            data.append(encode_msg(value[k]))
        elif k in value:
            data.append(b(builtin_str(value[k])))
    signature = b'&'.join(data)
    return sign == make_content_md5(signature)

if __name__ == '__main__':
    pass



# -*- coding: utf-8 -*-
from .compat import builtin_str


class UpYunServiceException(Exception):
    def __init__(self, request_id, status, msg, err):
        self.args = (request_id, status, msg, err)
        self.request_id = request_id
        self.status = status
        self.msg = msg
        self.err = err


class UpYunClientException(Exception):
    def __init__(self, msg):
        self.msg = builtin_str(msg)
        super(UpYunClientException, self).__init__(msg)



# -*- coding: utf-8 -*-

from .modules.sign import make_content_md5
from .modules.exception import UpYunServiceException, UpYunClientException
from .upyun import UpYun, ED_AUTO, ED_TELECOM, ED_CNC, ED_CTT,\
    __version__, verify_put_sign

__title__ = 'upyun'
__author__ = 'Monkey Zhang (timebug)'
__license__ = 'MIT License: http://www.opensource.org/licenses/mit-license.php'
__copyright__ = 'Copyright 2015 UPYUN'

__all__ = [
    'UpYun', 'UpYunServiceException', 'UpYunClientException',
    'ED_AUTO', 'ED_TELECOM', 'ED_CNC', 'ED_CTT', '__version__',
    'verify_put_sign', 'make_content_md5'
]



# -*- coding: utf-8 -*-
import os
import time
import math
import itertools
import threading
from multiprocessing.dummy import Pool as ThreadPool

from .modules.compat import urlencode, b
from .modules.exception import UpYunServiceException, UpYunClientException
from .modules.sign import make_policy, make_multi_signature, make_content_md5


class Multipart(object):
    def __init__(self, bucket, secret, endpoint, hp):
        self.bucket = bucket
        self.secret = secret
        self.hp = hp
        self.host = 'm0.api.upyun.com'
        self.uri = '/%s/' % bucket

    # --- public API
    def upload(self, key, value, block_size, expiration, **kwargs):
        lock = threading.Lock()
        expiration = expiration or 1800
        expiration = int(expiration + time.time())
        block_size = block_size or 1024*1024
        file_size = int(self.__get_size(value))
        block_size = self.__check_size(block_size)
        blocks = int(math.ceil(file_size * 1.0 / block_size)) or 1

        # - init upload
        content = self.__init_upload(key, value,
                                     file_size, blocks, expiration, **kwargs)
        if 'save_token' in content and 'token_secret' in content:
            save_token = content['save_token']
            token_secret = content['token_secret']
        else:
            raise UpYunServiceException(None, 503, 'Service unavailable',
                                        'Not enough response datas from api')
        status = self.__get_status(content)

        # - block item upload
        retry = 0
        pool = ThreadPool(4)
        while not self.__upload_success(status) and retry < 5:
            status_list = pool.map(lambda parms: self.__block_upload(*parms),
                                   zip(range(blocks), itertools.repeat(
                                       (status, value, file_size,
                                        block_size, expiration,
                                        save_token, token_secret, lock))
                                       ))
            status = self.__find_max_status(status_list)
            retry += 1
        pool.close()
        pool.join()

        # - end upload
        if self.__upload_success(status):
            return self.__end_upload(expiration, save_token, token_secret)
        else:
            UpYunServiceException(None, 500, 'Upload failed',
                                  'Failed to upload the whole '
                                  'file within retry times')

    # --- private API
    def __init_upload(self, key, value, file_size,
                      blocks, expiration, **kwargs):
        data = {'expiration': expiration,
                'file_blocks': blocks,
                'file_hash': make_content_md5(value),
                'file_size': file_size,
                'path': key,
                }

        data.update(kwargs)
        policy = make_policy(data)
        signature = make_multi_signature(data, self.secret)
        postdata = {'policy': policy, 'signature': signature}
        return self.__do_http_request(postdata)

    def __block_upload(self, index, parms):
        status, value, file_size, block_size, expiration,\
            save_token, token_secret, lock = parms
        # - if status[index] is already 1, skip it
        if status[index]:
            return status
        start_position = index * block_size
        if index == len(status) - 1:
            end_position = file_size
        else:
            end_position = start_position + block_size

        file_block = self.__read_block(value, start_position,
                                       end_position, lock)
        block_hash = make_content_md5(file_block)

        data = {'expiration': expiration, 'block_index': index,
                'block_hash': block_hash, 'save_token': save_token,
                }
        policy = make_policy(data)
        signature = make_multi_signature(data, token_secret)
        postdata = {'policy': policy,
                    'signature': signature,
                    'file': file_block,
                    }
        resp = self.hp.do_http_pipe('POST', self.host, self.uri,
                                    files=postdata)
        content = self.__handle_resp(resp)
        return self.__get_status(content)

    def __end_upload(self, expiration, save_token, token_secret):
        data = {'expiration': expiration, 'save_token': save_token}
        policy = make_policy(data)
        signature = make_multi_signature(data, token_secret)
        postdata = {'policy': policy, 'signature': signature}
        return self.__do_http_request(postdata)

    def __find_max_status(self, status_list):
        max_item = 0
        max_status = status_list[0]
        for status in status_list:
            if sum(status) > max_item:
                max_item = sum(status)
                max_status = status
        return max_status

    def __get_status(self, content):
        if 'status' in content and type(content['status']) == list:
            return content['status']
        else:
            raise UpYunServiceException(None, 503, 'Service unavailable',
                                        'Update status failed')

    def __do_http_request(self, value):
        headers = {'Content-Type': 'application/x-www-form-urlencoded'}

        value = urlencode(value)
        resp = self.hp.do_http_pipe('POST', self.host, self.uri,
                                    headers=headers, value=value)
        return self.__handle_resp(resp)

    def __handle_resp(self, resp):
        try:
            content = resp.json()
        except Exception as e:
            raise UpYunClientException(e)
        return content

    def __upload_success(self, status):
        return len(status) == sum(status)

    def __get_size(self, fileobj):
        try:
            if hasattr(fileobj, 'fileno'):
                return os.fstat(fileobj.fileno()).st_size
        except IOError:
            pass
        return len(fileobj.getvalue())

    def __check_size(self, block_size):
        if block_size > 5 * 1024 * 1024:
            block_size = 5 * 1024 * 1024
        if block_size < 100 * 1024:
            block_size = 100 * 1024
        return int(block_size)

    def __read_block(self, f, current_position, end_position, lock,
                     length=3*8192):
        data = []
        lock.acquire()
        while current_position < end_position:
            if (current_position + length) > end_position:
                length = end_position - current_position
            f.seek(current_position, 0)
            data.append(f.read(length))
            current_position += length
        lock.release()
        return b('').join(b(data))



# -*- coding: utf-8 -*-
import os
import time

from .modules.exception import UpYunClientException
from .modules.sign import make_content_md5, make_policy
from .modules.compat import b


class FormUpload(object):
    def __init__(self, bucket, secret, endpoint, hp):
        self.bucket = bucket
        self.secret = secret
        self.hp = hp
        self.host = endpoint
        self.uri = '/%s/' % bucket

    def upload(self, key, value, expiration, **kwargs):
        expiration = expiration or 1800
        expiration += int(time.time())

        data = {'bucket': self.bucket,
                'expiration': expiration,
                'save-key': key,
                }
        data.update(kwargs)
        policy = make_policy(data)
        signature = make_content_md5(policy + b('&') + b(self.secret))
        postdata = {'policy': policy,
                    'signature': signature,
                    'file': (os.path.basename(value.name), value),
                    }
        resp = self.hp.do_http_pipe('POST', self.host, self.uri,
                                    files=postdata)
        return self.__handle_resp(resp)

    def __handle_resp(self, resp):
        content = None
        try:
            content = resp.json()
        except Exception as e:
            raise UpYunClientException(e)
        return content



#! /usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import with_statement

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

import os
import sys
import upyun

if sys.argv[-1] == 'publish':
    os.system('python setup.py sdist upload')
    sys.exit()

setup(
    name='upyun',
    version=upyun.__version__,
    description='UpYun Storage SDK for Python',
    license='License :: OSI Approved :: MIT License',
    platforms='Platform Independent',
    author='Monkey Zhang (timebug)',
    author_email='timebug.info@gmail.com',
    url='https://github.com/upyun/python-sdk',
    packages=['upyun', 'upyun.modules'],
    keywords=['upyun', 'python', 'sdk'],
    install_requires=['requests>=2.4.3'],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
    ],
)



# -*- coding: utf-8 -*-

import hashlib
import base64
import json

from .compat import b, PY3, builtin_str, bytes, str
from .exception import UpYunClientException

DEFAULT_CHUNKSIZE = 8192


def make_content_md5(value, chunksize=DEFAULT_CHUNKSIZE):
    if hasattr(value, 'fileno'):
        md5 = hashlib.md5()
        for chunk in iter(lambda: value.read(chunksize), b''):
            md5.update(chunk)
        value.seek(0)
        return md5.hexdigest()
    elif isinstance(value, bytes) or (not PY3 and
                                      isinstance(value, builtin_str)):
        return hashlib.md5(value).hexdigest()
    else:
        raise UpYunClientException('object type error')


def decode_msg(msg):
    if isinstance(msg, bytes):
        msg = msg.decode('utf-8')
    return msg


def encode_msg(msg):
    if isinstance(msg, str):
        msg = msg.encode('utf-8')
    return msg


def make_policy(data):
    policy = json.dumps(data)
    return base64.b64encode(b(policy))


def make_rest_signature(bucket, username, password,
                        method, uri, date, length):
    if method:
        signstr = '&'.join([method, uri, date, str(length), password])
        signature = hashlib.md5(b(signstr)).hexdigest()
        return 'UpYun %s:%s' % (username, signature)

    else:
        signstr = '&'.join([uri, bucket, date, password])
        signature = hashlib.md5(b(signstr)).hexdigest()
        return 'UpYun %s:%s:%s' % (bucket, username, signature)


def make_multi_signature(data, secret):
    list_meta = sorted(data.items(), key=lambda d: d[0])
    signature = ''.join(map(lambda kv: '%s%s' %
                        (kv[0], str(kv[1])), list_meta))
    signature += secret
    return make_content_md5(b(signature))


def make_av_signature(data, operator, password):
    assert isinstance(data, dict)
    signature = ''.join(map(lambda kv: '%s%s' %
                        (kv[0],
                         kv[1] if type(kv[1]) != list else ''.join(kv[1])),
                        sorted(data.items())))
    signature = '%s%s%s' % (operator, signature, password)
    return make_content_md5(b(signature))



# -*- coding: utf-8 -*-
from .exception import UpYunClientException


# a decoration to check if example has members
def has_object(obj_name):
    def deco(func):
        def _(self, *a, **kw):
            if not hasattr(self, obj_name):
                msg = 'Class Upyun dont have an attr called %s' % obj_name
                raise UpYunClientException(msg)
            return func(self, *a, **kw)
        return _
    return deco
